#!/usr/bin/python2.7
# -*- coding: utf-8 -*-
#
#  OX for UCS
#  listener module: handling groups
#
# Copyright (C) 2004-2017 Univention GmbH <http://www.univention.de/>
# and
#     iKu Systems & Services GmbH & Co. KG (http://www.iku-systems.de/)
#
# All rights reserved.
#
# The source code of this program is made available
# under the terms of the GNU Affero General Public License version 3
# (GNU AGPL V3) as published by the Free Software Foundation.
#
# Binary versions of this program provided by Univention to you as
# well as other copyrighted, protected or trademarked materials like
# Logos, graphics, fonts, specific documentations and configurations,
# cryptographic keys etc. are subject to a license agreement between
# you and Univention and not subject to the GNU AGPL V3.
#
# In the case you use this program under the terms of the GNU AGPL V3,
# the program is provided in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License with the Debian GNU/Linux or Univention distribution in file
# /usr/share/common-licenses/AGPL-3; if not, see
# <http://www.gnu.org/licenses/>.

__package__ = ''  # workaround for PEP 366
import os
import listener
from univention.ox.oxlistener import OxListener
from univention.ox.listener_tools import ListenerQueue, get_credentials, ox_list_user, ox_list_group, ox_call, listener_is_deactivated

name = 'ox-groups'
description = 'Group settings for Open-Xchange'
filter = '(objectClass=posixGroup)'
attributes = []


# language
os.environ["LANG"] = listener.configRegistry.get("locale/default", "").split(':')[0]


class OXGroup(object):
	def __init__(self, groupname, context_ids):  # type: (str, List[str]) -> None
		self.groupname = groupname
		self.context_ids = context_ids
		self.displayname = dict()
		self.members = dict()

		for context_id in self.context_ids:
			members = self.retrieve_members(context_id)
			if members is not None:
				self.members[context_id] = members

	def retrieve_members(self, context_id):  # type: (str) -> Dict[str, List[str]]
		admuser, admpass = OxListener.get_credentials(context_id)
		for oxgroupelement in ox_list_group(admuser, admpass, context_id, searchpattern=self.groupname):
			if oxgroupelement[OxListener.CSV_USER_ROW_GROUPNAME] == self.groupname:
				self.displayname[context_id] = oxgroupelement[OxListener.CSV_USER_ROW_DISPLAYNAME]
				return oxgroupelement[OxListener.CSV_USER_ROW_MEMBERS].split(',')


class OXGroupListener(OxListener):
	def __init__(self, name, old, new):  # type: (str, Dict[str, List[str]], Dict[str, List[str]]) -> None
		super(OXGroupListener, self).__init__(name, old, new)

		self.old_context_ids = list()
		self.new_context_ids = list()
		self.groupname = self.get_groupname()
		self.logger.qualifier = self.groupname
		self.ox_group = None
		self.new_members = dict()
		self.old_members = dict()
		self._member_cache = dict()
		self._ox_userid_cache = dict()

	def create_group(self, context_id):  # type: (str) -> None
		self.logger.process('Creating group "{}" in context {}...'.format(self.groupname, context_id))

		if not self.new_members:
			self.new_members = self.get_ox_members(self.new)
			self.logger.debug('OX users NEW by context: {}'.format(self.new_members))

		admuser, admpass = self.get_credentials(context_id)
		command = [
			'creategroup',
			'--displayname={}'.format(self.get_description()),
		]
		if self.new_members.get(context_id):
			command.append('--addmembers={}'.format(','.join(self.new_members[context_id].values())))

		command.extend([
			'--adminuser={}'.format(admuser),
			'--adminpass={}'.format(admpass),
			'--contextid={}'.format(context_id),
			'--name={}'.format(self.groupname),
		])
		ox_call(command, logger=self.logger)

	def get_groupname(self):  # type: () -> str
		if self.new.get('cn'):
			return self.new['cn'][0]
		elif self.old.get('cn'):
			return self.old['cn'][0]
		else:
			return ''

	def get_description(self):  # type: () -> str
		res = self.new.get('description', [self.groupname])[0]
		return res[:128]  # OX length limit (Bug #35820)

	def get_ox_members(self, old_or_new):  # type: (str, Dict[str, List[str]]) -> Dict[str, Dict[str, Dict[str, str]]]
		"""
		Get all usernames and user-ids from all contexts for this groups members.

		:param old_or_new: dict: listener new/old
		:return: dict: {context id: {username: ox_user_id, ..}, ..}
		Example: {'280': {'test19': '31', 'user28': '3'}, '12': {'user32': '39'}, '10': {'test30': '179', 'user31': '162'}}
		"""
		ox_users = dict()
		try:
			member_dns = old_or_new['uniqueMember']
		except (KeyError, TypeError):
			return ox_users

		# collect context IDs and usernames of members
		for dn in member_dns:
			if dn not in self._member_cache:
				udm_user = self.get_udm_user(dn)
				if udm_user.info.get('isOxUser', '').lower() == 'ok':
					user_context = udm_user.info.get('oxContext', self.get_default_context_id())
					user_name = udm_user.info['username']
					self._member_cache[dn] = dict(context=user_context, username=user_name)
				else:
					continue

			context = self._member_cache[dn]['context']
			username = self._member_cache[dn]['username']
			ox_users.setdefault(context, dict())[username] = None

		# add OX user IDs
		for context_id, usernames in ox_users.items():
			if context_id not in self._ox_userid_cache:
				admuser, admpass = self.get_credentials(context_id)
				for row in ox_list_user(admuser, admpass, context_id):
					row_username = row[self.CSV_USER_ROW_USERNAME]
					ox_user_id = row[self.CSV_USER_ROW_UID]
					self._ox_userid_cache.setdefault(context_id, dict())[row_username] = ox_user_id
			for username in usernames.keys():
				try:
					usernames[username] = self._ox_userid_cache[context_id][username]
				except KeyError:
					# user has been deleted in the meantime
					del usernames[username]
		return ox_users

	def modify_group(self, context_id):  # type: (str) -> None
		self.logger.process('Modifying group "{}" in context {}...'.format(self.groupname, context_id))

		if not self.new_members:
			self.new_members = self.get_ox_members(self.new)
			self.logger.debug('OX group members NEW by context: {}'.format(self.new_members))
		if not self.old_members:
			self.old_members = self.get_ox_members(self.old)
			self.logger.debug('OX group members OLD by context: {}'.format(self.old_members))

		add_members = set(self.new_members.get(context_id, {}).values()) - set(self.old_members.get(context_id, {}).values())
		add_members.difference_update(set(self.ox_group.members.get(context_id, [])))  # don't try to add existing members
		rm_members = set(self.old_members.get(context_id, {}).values()) - set(self.new_members.get(context_id, {}).values())

		if not add_members and not rm_members and self.ox_group.displayname[context_id] == self.get_description():
			self.logger.process('No changes for OX group "{}" in context {}.'.format(self.groupname, context_id))
			return

		admuser, admpass = self.get_credentials(context_id)
		command = [
			'changegroup',
			'--displayname={}'.format(self.get_description()),
		]
		if rm_members:
			command.append("--removemembers=%s" % ','.join(rm_members))
		if add_members:
			command.append("--addmembers=%s" % ','.join(add_members))
		command.extend([
			'--adminuser={}'.format(admuser),
			'--adminpass={}'.format(admpass),
			'--contextid={}'.format(context_id),
			'--name={}'.format(self.groupname),
		])
		ox_call(command, logger=self.logger)

	@property
	def new_is_ox_group(self):  # type: () -> bool
		return self.new.get('isOxGroup', [''])[0].lower() == 'ok'

	@property
	def old_is_ox_group(self):  # type: () -> bool
		return self.old.get('isOxGroup', [''])[0].lower() == 'ok'

	def run(self):  # type: () -> None
		# ignore groups with name "users" (Bug #35821)
		if self.groupname.lower() == "users":
			self.logger.process('Ignoring group "{}".'.format(self.groupname))
			return

		if self.old:
			self.old_context_ids = self.get_context_ids_from_group(self.old)
			self.logger.process('Context IDs of group "{}" in old: {}'.format(self.groupname, self.old_context_ids))
		else:
			self.old_context_ids = list()
		if self.new:
			self.new_context_ids = self.get_context_ids_from_group(self.new)
			self.logger.process('Context IDs of group "{}" in new: {}'.format(self.groupname, self.new_context_ids))
		else:
			self.new_context_ids = list()

		if self.new_is_ox_group:
			# remove group from missing contexts
			for context_id in set(self.old_context_ids) - set(self.new_context_ids):
				self.remove_group(self.groupname, context_id)

			# modify or create contexts
			self.ox_group = OXGroup(self.groupname, self.new_context_ids)
			self.logger.debug('Current members of OX group "{}": {}'.format(self.groupname, self.ox_group.members))

			for context_id in self.new_context_ids:
				if context_id in self.ox_group.members:
					# group exists in context
					self.modify_group(context_id)
				else:
					self.create_group(context_id)
			return

		# remove group from all contexts
		if self.old_is_ox_group and not self.new_is_ox_group:
			for context_id in self.old_context_ids:
				self.remove_group(self.groupname, context_id)
			return


@ListenerQueue(name, stop_on_error=False)
def handler(dn, new, old, action=None):  # type: (Dict[str, List[str]], Dict[str, List[str]], Optional[str]) -> None
	if listener_is_deactivated(name):
		return

	ox_group_listener = OXGroupListener(name, old, new)
	ox_group_listener.run()


if __name__ == '__main__':
	handler.process_pending()
